#cython: language_level=3
# Copyright (c) 2022 Intel Corporation.
# The following code is Tool Software as licensed in the Edge Software Master Agreement.


import os
import re
import time
import uuid
import socket
import json
import configparser
import getpass
import urllib
import random
import string
import requests
import shlex
import sys
import subprocess
import ipaddress
from termcolor import colored,cprint
from click import progressbar
from ruamel.yaml import YAML
from pathlib import Path
from datetime import datetime
from multiprocessing import Process, Value
from subprocess import run, PIPE, check_output, DEVNULL, STDOUT, CalledProcessError

from esb_common.logger import Logger
from esb_common import locale
from esb_common.util import Utility, Status


MODULE = "esm"
CURRENT_DIR = os.getcwd()

STACK = "esm"
ZIP_FILE = "setup.zip"

# Services started by application
SERVICES = [STACK + "_portainer",
            STACK + "_grafana",
            STACK + "_edge-software-manager",
            STACK + "_spice-web",
            STACK + "_esm-server",
            STACK + "_vm-server",
            STACK + "_vm-server_db",
            STACK + "_influxdb",
            STACK + "_registry",
            STACK + "_portainer_edge_agent"]

# Standalone container that are mandatory
SA_CONTAINER = [STACK + "-collectd"]

AMT_CONTAINER = [STACK + "_mps_1",
                 STACK + "_rps_1",
                 STACK + "_webui_1",
                 STACK + "_vault_1",
                 STACK + "_rpsdb_1"]

# Standalone containers that are not mandatory,
# but need to be removed on cleanup
OTHER_CONTAINERS = ["emt-recipe-installer",
                    "emt-recipe-uninstaller",
                    STACK + "-vm-agent",
                    "esc"+"_containerized-vm",]

# Directories or files present in host system that needs to be
# deleted on uninstallation
HOST_FILE_TO_REM = ["/opt/esm", "/opt/esc", "/etc/esm"]

DOCKER_VOLUMES = [STACK + "_portainer_data",
                  STACK + "_grafana",
                  STACK + "_portainer_agent_data",
                  STACK + "_vm-server_db_data",
                  STACK + "_influxdb-data",
                  STACK + "_registry-data",
                  STACK + "_esm-server_db_migrations_data",
                  STACK + "_esm-server",
                  STACK + "_private-volume",
                  STACK + "_app-volume",
                  STACK + "_gotify-server_db_data"]

DOCKER_SECRETS = {"influxdb-auth": "influxdb-auth",
                  "server.key": "server.key",
                  "server.crt": "server.crt",
                  "vmserver-key": "vmserver-key",
                  "esm-conf": "esm-conf",
                  "gotify_app_token": "gotify_app_token",
                  "gotify_client_token": "gotify_client_token",
                  }

# Images to be pushed to local registry
AGENT_IMAGES = ["esc_collectd:latest",
                "esc_vm-agent:latest",
                "esc_containerized-vm:latest",
                "esc_kernel-builder:latest"]

# Additional images built by setup
ADDITIONAL_IMAGES = ["uosbuilder:latest", "builder-dnsmasq:latest",
                     "builder-git:latest", "builder/dyninit:v1.0",
                     "builder-wget:latest", "builder-aws-cli:latest",
                     "alpine/kernel:v3.9", "ubuntu:20.04", "ubuntu:18.04",
                     "influxdb:1.8", "nginx:1.19.0", "",
                     "portainer/portainer:1.24.0", "grafana/grafana:6.3.3",
                     "postgres:12.3-alpine", "registry:2", "portainer/agent:1.6.0",
                     "linuxkit/rngd:v0.6", "debian:buster", "python:3.7.8-alpine3.12",
                     "python:3.7.8", "alpine:3.12", "alpine:3.9",
                     "linuxkit/go-compile:b1446b2ba407225011f97ae1dba0f512ae7f9b84",
                     "docker:19.03.0-dind", "docker:18.06.3-ce-dind", "datadog/squid:latest",
                     "nginx:1.14.2-alpine",
                     "linuxkit/runc:a81d48a1568f41b1c2e048fe017dbd88c6a4bdcc",
                     "linuxkit/init:4448c4b6d4308244160b71c423bc9df32bc180db",
                     "linuxkit/containerd:56e03cfa92d75d6eb5a7fc44742e50f427ba29a3",
                     "linuxkit/sshd:c4bc89cf0d66733c923ab9cb46198b599eb99320",
                     "linuxkit/sysfs:v0.6",
                     "linuxkit/dhcpcd:v0.6",
                     "linuxkit/ca-certificates:v0.6",
                     "linuxkit/modprobe:v0.6",
                     "linuxkit/openntpd:v0.6",
                     "linuxkit/sysctl:v0.6",
                     "linuxkit/getty:2eb742cd7a68e14cf50577c02f30147bc406e478",
                     "hello-world:latest",
                     "mps:latest",
                     "rps:latest",
                     "vault:latest",
                     "postgres:latest",
                     "webui:latest"]
REGISTRY_PORT = "5055"
RETRIES = 20
VMSERVER_TOKEN = ""
SERVER_HOSTNAMES = ["localhost", "portainer", "grafana", "esm-server",
                    "vm-server", "registry", "influxdb", "gotify"]

DOCKER_PROXY_DIR = "/etc/systemd/system/docker.service.d"
HTTP_PROXY_FILE = "http-proxy.conf"
HTTPS_PROXY_FILE = "https-proxy.conf"

BUILD_IMAGE_PROC = 5
CA_BUNDLE = "/etc/ssl/certs/ca-certificates.crt"

# HTTP Response Codes
SUCCESS = 200
UNPROCESSABLE_ENTITY = 422
CREATED = 201

AMT_REPO_TAG = "v1.1.0"
AMT_REPO = "https://github.com/open-amt-cloud-toolkit/open-amt-cloud-toolkit.git"
amt_dir = "open-amt-cloud-toolkit"
sys_python_ver = sys.version.split()[0][0:3]
PYTHON_VER=['3.8','3.6']


class EdgeSoftwareManagerException(Exception):

    """
    Exception handling class
    """
    pass


class EdgeSoftwareManager(object):
    """
    Main class for the module EdgeSoftwareManager
    """

    def __init__(self, module_path, output_path):
        self.module = MODULE
        self.output_path = output_path
        self.module_path = module_path
        log_file = "{0}/install.log".format(self.output_path)
        if not os.path.isdir(self.output_path):
            os.mkdir(self.output_path)
        self.log = Logger(log_file)
        self.pbar = None
        self.str_len = 50
        self.ip_address = self.get_ip_address()
        self.hostname = socket.gethostname()
        self.base_url = "https://{0}:3002".format(self.hostname)
        self.portainer_url = "{0}/portainer".format(self.base_url)
        self.vmserver_url = "{0}/vmserver".format(self.base_url)
        self.reinstallation = False
        self.user = None
        self.password = None
        self.jwt = None
        self.vmkey = None
        self.log_file = []
        self.package_label = None
        self.package_version = None
        self.agent_uuid = None
        self.pbar = None
        self.gotify_username = "admin"
        self.username = "admin"
        self.password = None
        self.setup_start_time = datetime.now()
        self.down_interface = ""

    def install(self):
        """
        Method to call main_install and main_post_install with exception handling
        :return: status :type: bool
        """
        status = True
        try:
            self.main_install()
            self.main_post_install()
            self.log.info("Successfully installed {0}".format(self.module))
        except EdgeSoftwareManagerException as error:
            if error.args[0]:
                if isinstance(error.args[0], bytes):
                    self.log.error(error.args[0].decode('utf-8'))
                else:
                    self.log.error(error.args[0])
            if error.args[1]:
                self.log.error(error.args[1])
                print("{0} {1}".format(colored("ERROR:", "red"), error.args[1]))
            if error.args[2]:
                self.log.error(error.args[2])
                print("{0} {1}".format(colored("Failed command:", "red"),
                                       error.args[2]))
            print("{0} Cleaning up ...".format(colored("Setup Failed.", "red")))
            self.remove_stack()
            status = False
        except (KeyboardInterrupt, PermissionError):
            print("{0} {1}".format(colored("WARNING:", "yellow"),
                                   locale.USER_ABORTED))
            self.log.warn(locale.USER_ABORTED)
            print("Cleaning up ...")
            self.remove_stack()
            status = False
        except Exception as err:
            print("{0} {1}".format(colored("ERROR:", "red"), err))
            self.log.error(err)
            status = False
            print("{0} Cleaning up ...".format(colored("Setup Failed.", "red")))
            self.remove_stack()
        finally:
            self.cleanup()
            os.chdir(CURRENT_DIR)
        return status

    def run_command(self, command, message="", input=None, stdout=PIPE,
                    stderr=PIPE, mandatory=True, get_output=False):
        final_command = shlex.split(command)
        status = run(final_command, input=input, stdout=stdout, stderr=stderr)
        if get_output:
            return status
        if status.returncode:
            error = ""
            if stderr == PIPE and stdout == PIPE:
                error = status.stdout + status.stderr
            if mandatory:
                raise EdgeSoftwareManagerException(error,
                                                   message,
                                                   command)
            else:
                if error:
                    self.log.warn(error.decode('utf-8'))
                self.log.warn(message)
                self.log.warn(command)
            return False
        # self.pbar.update(1)
        return True

    def check_command_output(self, command, input=None):
        final_command = shlex.split(command)
        return check_output(final_command, input=input)

    def check_docker(self):
        """
        Check if docker is installed or not
        :return: Bool value specifying if docker is installed properly
        """
        self.log.info(locale.VERIFY_EXISTING_SETUP)
        with progressbar(length=3, fill_char=".", empty_char=" ", width=50,
                         label="Verifying the Dependencies".ljust(
                             self.str_len, " "),
                         show_eta=True, show_percent=True,
                         bar_template="%(label)s [%(bar)s] %(info)s") as pbar:
            self.log.info(locale.VERIFY_EXISTING_SETUP)
            time.sleep(1)
            pbar.update(1)
            self.run_command("docker -v", message="Docker is not Installed")
            status = self.run_command("systemctl is-active docker", get_output=True)
            if status.returncode == 3:
                message = "Docker engine is not running!"
                raise EdgeSoftwareManagerException("", message, "")
            pbar.update(1)
            self.run_command("sudo docker rmi -f hello-world")
            message = "Docker CE {0}, but failed to run docker hello-world".format(
                locale.ALREADY_INSTALLED) + ". " + locale.INCORRECT_PROXY_ENTRY
            self.run_command("sudo docker run --rm hello-world", message)
            pbar.update(1)
            self.log.info("Docker " + locale.VERIFIED)

    def init_swarm(self):
        """
        Initialize the docker swarm
        :return: None
        """
        self.log.info("Initializing docker swarm")
        command = f"sudo docker swarm init --advertise-addr {self.ip_address}"
        status = self.run_command(command, get_output=True)
        if status.returncode:
            error = status.stderr.decode("utf-8")
            if "This node is already part of a swarm" not in error:
                raise EdgeSoftwareManagerException(
                    status.stdout + status.stderr,
                    "Failed to initialise the swarm",
                    command)

    def init_admin(self, ca_path, username, password):
        """
        Initialises the administrator account in portainer
        :param ca_path: Path to CA certificate
        :param username: Username of admin account
        :param password: Password for admin account
        :return:
        """
        try:
            if self.reinstallation:
                return Status.SUCCESS
            headers = {"Content-Type": "application/json"}
            data = {"Username": username, "Password": password}
            res = requests.post("{0}/api/users/admin/init".format(self.portainer_url),
                                headers=headers,
                                json=data,
                                verify=ca_path)
            if res.status_code != SUCCESS:
                return res
            else:
                return Status.SUCCESS
        except Exception as error:
            self.log.info(error)
            self.log.info("Trying to initialise admin account. Waiting for "
                          "server to start...")
            return Status.FAILED

    def update_service(self, service):
        """
        Update the docker services
        """
        self.log.info(f"Updating {service}")
        command = f"sudo docker service update {service} --force"
        self.run_command(command, f"Failed to update the service {service}",
                         mandatory=False)

    def deploy_gotify(self, gotify_generated_pasword):
        """
        Deploy the gotify servers using docker-compose-gotify.yml file
        :return:
        """
        original_dir = os.getcwd()
        try:
            self.log.info("Deploying Gotify")
            # If ESM is being re-installed, skip the Gotify installation
            if self.reinstallation:
                return

            os.chdir(os.path.join(self.module_path, "setup"))

            command = "sudo GOTIFY_GENERATED_PASSWORD={0} GOTIFY_DB_PASSWORD={0} " \
                      "docker stack deploy -c docker-compose-gotify.yml {1} " \
                      "--with-registry-auth".format(
                      gotify_generated_pasword, STACK)
            self.run_command(command, "Failed to deploy Gotify")

        except Exception as e:
            raise EdgeSoftwareManagerException("",
                "Failed to deploy Gotify. Error = {}".format(e), "")
        finally:
            # Change back to the original directory
            os.chdir(original_dir)

    def register_app(self, ca_path, gotify_generated_pasword):

        register_url = "https://" + self.ip_address + ":5052/application"
        headers = {'Content-Type': 'application/json'}
        payload = {"description": "ESM Server APP", "name": "ESM"}
        response = requests.post(register_url,
                                 auth=(
                                     self.gotify_username,
                                     gotify_generated_pasword),
                                 headers=headers, json=payload, verify=ca_path)
        if response.status_code != SUCCESS:
            self.log.error("Gotify Application Registration Failed!")
            return None

        json_response = json.loads(response.text)
        application_token = json_response["token"]
        self.log.info("Gotify Application Registration Successful")
        return application_token

    def register_client(self, ca_path, gotify_generated_pasword):

        register_url = "https://" + self.ip_address + ":5052/client"
        headers = {'Content-Type': 'application/json'}
        payload = {"description": "ESM Client", "name": "ESM_Client"}
        response = requests.post(register_url,
                                 auth=( self.gotify_username,
                                     gotify_generated_pasword),
                                 headers=headers, json=payload, verify=ca_path)
        if response.status_code != SUCCESS:
            self.log.error("Gotify Client Registration Failed!")
            return None

        json_response = json.loads(response.text)
        client_token = json_response["token"]
        self.log.info("Gotify Client Registration Successful")
        return client_token

    def gotify_create_token_secrets(self, gotify_app_token, gotify_client_token):
        if gotify_app_token is None or gotify_client_token is None:
            self.log.error("Failed to register application & client on Gotify")
            raise EdgeSoftwareManagerException("",
                "Failed to register application & client on Gotify. "
                "Gotify app/client token empty", "")

        command = f"sudo docker secret create {DOCKER_SECRETS['gotify_app_token']} -"
        self.run_command(command,
                         "Failed to generate gotify app token secret",
                         input=gotify_app_token.encode())
        command = f"sudo docker secret create {DOCKER_SECRETS['gotify_client_token']} -"
        self.run_command(command,
                         "Failed to generate Gotify client secret",
                         input=gotify_client_token.encode())

    def deploy_esm_ui(self):
        """
        Deploy the esm servers using docker-compose-ui.yml file
        :return:
        """
        os.chdir(os.path.join(self.module_path, "setup"))
        self.log.info("Deploying Edge Software Configurator UI")
        command = "sudo -E HTTPSERVER={0} BASE_URL={0} DOCKER_REGISTRY={0} " \
                  "DOCKER_REGISTRY_PORT={1} HOST_IP={2} docker stack deploy " \
                  "-c docker-compose-ui.yml {3} --with-registry-auth".format(
                  self.hostname, REGISTRY_PORT, self.ip_address, STACK)

        status = self.run_command(command, get_output=True)
        if status.returncode:
            if not self.reinstallation:
                raise EdgeSoftwareManagerException(status.stdout + status.stderr,
                                                   "Failed to deploy Edge Software Configurator",
                                                   command)

        if self.reinstallation:
            services = [STACK + "_edge-software-manager",
                        STACK + "_spice-web",
                        STACK + "_esm-server",
                        STACK + "_vm-server"]
            for service in services:
                self.update_service(service)

    def get_token(self, ca_path, username, password):
        """
        Get JWT token by authenticating the user
        :param ca_path: Path to CA certificate
        :param username: Username of user
        :param password: Password
        :return: JWT token on success else response of the auth request
        """
        self.log.info("Authenticating User")
        headers = {"Content-Type": "application/json"}
        data = {"Username": username, "Password": password}
        res = requests.post("{0}/api/auth".format(self.portainer_url),
                            headers=headers,
                            json=data,
                            verify=ca_path)
        if res.status_code != SUCCESS:
            return res
        else:
            res_data = json.loads(res.text)
            return res_data["jwt"]

    def get_ip_address(self):
        """
        Get the IP address of the Build node
        :return: None
        """
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect(("8.8.8.8", 80))
            ip_address = sock.getsockname()[0]
            sock.close()
        except Exception:
            raise EdgeSoftwareManagerException("",
                                               "Failed to get the IP address",
                                               "")
        return ip_address

    def create_endpoint(self, jwt_token, ca_path):
        """
        Create an endpoint in portainer.
        :param jwt_token: JWT token
        :param ca_path: Path to A certificate
        :return: Edge key, Endpoint Id
        """
        self.log.info("Creating the Edge Software Configurator endpoint")
        headers = {"Authorization": "Bearer {0}".format(jwt_token)}
        data = {"Name": self.hostname, "EndpointType": "4",
                "URL": self.portainer_url}
        res = requests.post("{0}/api/endpoints".format(self.portainer_url),
                            headers=headers,
                            data=data,
                            verify=ca_path)
        if res.status_code != SUCCESS:
            return res
        else:
            res_data = json.loads(res.text)
            return res_data["EdgeKey"], res_data["Id"]

    def collectd_config(self):
        """
        Configure collectd config file with IP address of build node where
        Influxdb is running
        :return: None
        """
        self.log.info("Configuring Collectd")
        username = 'admin'
        password = 'admin'
        auth_credential = str(username) + ": " + str(password) + "\n"
        if self.reinstallation:
            command = f"sudo docker service rm {STACK}_influxdb"
            self.run_command(command,
                             message="Failed to remove influxdb",
                             input=auth_credential.encode(), mandatory=False)

            time.sleep(3)
            command = f"sudo docker secret rm {DOCKER_SECRETS['influxdb-auth']}"
            self.run_command(command,
                             "Failed to remove influxdb",
                             input=auth_credential.encode(), mandatory=False)

        command = f"sudo docker secret create {DOCKER_SECRETS['influxdb-auth']} -"
        self.run_command(command,
                         "Failed to configure Influxdb",
                         input=auth_credential.encode())

    def update_server_secret(self, username, password, edge_id, endpoint_id=1):
        """
        Update the esm-conf secret if Endpoint ID is other than one
        :return: None
        """
        self.log.info("Updating Edge Software Configurator server configs")
        message = "Unable to update esm-server secret"
        command = "sudo docker service update --secret-rm {} esm_esm-server".format(
            DOCKER_SECRETS["esm-conf"])
        self.run_command(command, message)
        command = "sudo docker secret rm {}".format(DOCKER_SECRETS["esm-conf"])
        self.run_command(command, message)
        self.generate_config_file(username, password, edge_id, endpoint_id)
        command = "sudo docker service update --secret-add source={}," \
                  "target=/etc/esm/esm.conf,mode=0400 esm_esm-server".format(DOCKER_SECRETS["esm-conf"])
        self.run_command(command, message)

    def get_edge_key(self, ca_path):
        """
        Get the Edge agent key
        """
        endpoint_id = 1
        url = f"{self.portainer_url}/api/endpoints/{endpoint_id}"
        headers = {"Authorization": f"Bearer {self.jwt}"}
        res = requests.get(url, headers=headers, verify=ca_path)
        if res.status_code != SUCCESS:
            self.log.error("Error occured while getting edge key")
            self.log.error(f"{res.status_code}: {res.text}")
            return res
        else:
            res_data = json.loads(res.text)
            key = res_data.get("EdgeKey", "")
            edgeid = res_data.get("EdgeID", "")
            return key, edgeid

    def deploy_esm_agents(self, pbar, username, password):
        """
        Deploy the ESM agents
        :param pbar: progress bar object
        :param username: username of portainer admin account
        :param password: password
        :return: None
        """
        agent_env = {}
        self.log.info("Deploying Edge Software Configurator agents")
        agent_env["EDGE"] = "1"
        agent_env["EDGE_ID"] = self.agent_uuid
        result = False
        if self.reinstallation:
            ca_path = CA_BUNDLE
        else:
            ca_path = os.path.join(self.module_path, "setup/certs/esm-ca.crt")
        status = self.init_admin(ca_path, username, password)
        if status != Status.SUCCESS:
            for i in range(RETRIES):
                status = self.init_admin(ca_path, username, password)
                if status == Status.SUCCESS:
                    result = True
                    break
                time.sleep(5)
                pbar.update(1)
            if not result:
                if not isinstance(status, int):
                    raise EdgeSoftwareManagerException(status.text,
                                                       "Failed to initialise Edge Software Configurator",
                                                       "")
                else:
                    raise EdgeSoftwareManagerException("",
                                                       "Failed to initialise Edge Software Configurator",
                                                       "")
        agent_env["docker_registry"] = "{0}:{1}".format(self.hostname,
                                                REGISTRY_PORT)
        if self.reinstallation:
            os.environ["no_proxy"] = self.ip_address
            self.user = getpass.getuser()
            for i in range(3):
                self.password = getpass.getpass(prompt=f"password for {self.user}: ")
                jwt, vmkey = self.esm_login()
                if jwt:
                    if jwt == "Unauthorized":
                        print("Sorry, try again.")
                        continue
                    self.jwt = jwt
                    self.vmkey = vmkey
                    break
                else:
                    print("Failed to authenticate User. Reinstallation Failed.")
            if not self.jwt:
                return Status.FAILED

            status = self.get_edge_key(ca_path)
            if isinstance(status, tuple):
                agent_env["EDGE_KEY"] = status[0]
                agent_env["EDGE_ID"] = status[1]
            else:
                raise EdgeSoftwareManagerException("",
                                                   "Failed to get the agent Edge Key",
                                                   "")

        else:
            status = self.get_token(ca_path, username, password)
            if isinstance(status, str):
                jwt_token = status
            else:
                raise EdgeSoftwareManagerException(status.text,
                                                   "Failed to get the endpoint token",
                                                   "")
            status = self.create_endpoint(jwt_token, ca_path)
            if isinstance(status, tuple):
                agent_env["EDGE_KEY"] = status[0]
                endpoint_id = status[1]
                if endpoint_id != 1:
                    self.update_server_secret(username, password,
                                              agent_env["EDGE_ID"], endpoint_id)
            else:
                raise EdgeSoftwareManagerException(status.text,
                                                   "Failed to create the endpoint",
                                                   "")
        self.agent_uuid = agent_env["EDGE_ID"]
        command = "sudo EDGE={0} EDGE_ID={1} EDGE_KEY={2} " \
                  " docker stack deploy -c docker-compose-agent.yml {3} " \
                  "--with-registry-auth" \
                .format(agent_env["EDGE"], agent_env["EDGE_ID"],
                    agent_env["EDGE_KEY"], STACK)
        self.run_command(command, "Failed to deploy the ESM Edge Agents")
        self.log.info("Deployed Edge Software Configurator agent")
        command = "sudo EDGE_ID={0} HOST_IP={1} docker-compose -p {2} " \
                  "up -d".format(agent_env["EDGE_ID"], self.ip_address, STACK)
        self.run_command(command, "Failed to deploy the Edge Software Configurator Agents")
        self.log.info("Deployed Telemetry agents")

    def deploy_amt_setup(self):
        """
        Deploy the MPS and RPS Servers
        :return: None
        """
        self.log.info("Deploying AMT setup")
        os.chdir(os.path.join(self.module_path, "setup/open-amt-cloud-toolkit"))
        command = f"sudo  docker-compose -p esm -f ./docker-compose.yml up --build -d"
        status = self.run_command(command, "Failed to deploy the AMT Microservices",
                                  mandatory=False)
        if not status:
            if not os.path.exists("/opt/amt-cert"):
                self.run_command("sudo mkdir /opt/amt-cert",
                                 "Failed to create directory", mandatory=False)

        os.chdir(CURRENT_DIR)

    def update_registry(self):
        """
        Update the Local registry with the required images
        :return: None
        """
        self.log.info("Updating Edge Software Configurator docker registry")
        # registry = "{0}:{1}".format(self.ip_address, REGISTRY_PORT)
        # changing ip to hostname
        registry = "{0}:{1}".format(self.hostname, REGISTRY_PORT)
        for image in AGENT_IMAGES:
            command = "sudo docker tag {0} {1}/{0}".format(image, registry)
            self.run_command(command, "Failed to Tag docker image {0}".format(image))
            command = "sudo docker push {1}/{0}".format(image, registry)
            message = "Failed to push the docker image {0} to local " \
                      "registry".format(image)
            self.run_command(command, message)

    def generate_certificates(self):
        """
        Generate CA certificate.
        User CA certificate to sign teh server certificates
        :return: None
        """
        self.log.info("Generating server certificates")
        root_cnf = "[root_ca]\n" \
                "basicConstraints = critical,CA:TRUE,pathlen:1\n" \
                "keyUsage = critical, nonRepudiation, cRLSign, keyCertSign\n" \
                "subjectKeyIdentifier=hash\n "

        server_cnf = "[server]\n" \
                    "authorityKeyIdentifier=keyid,issuer\n" \
                    "basicConstraints = critical,CA:FALSE\n" \
                    "extendedKeyUsage=serverAuth\n" \
                    "keyUsage = critical, digitalSignature, keyEncipherment\n" \
                    "subjectKeyIdentifier=hash\n" \
                    "subjectAltName = @alt_names"

        certs_path = os.path.join(self.module_path, "setup/certs")
        if not os.path.isdir(certs_path):
            os.mkdir(certs_path)
        os.chdir(certs_path)

        # Gernerate CA Key
        command = "openssl genrsa -out esm-ca.key 4096"
        self.run_command(command, "Failed to generate CA key")

        # Generate certificate signing request using  CA key
        command = "openssl req -new -key esm-ca.key -out esm-ca.csr -sha256 " \
                  "-subj '/O=Intel Corp/CN=edgesoftwaremanager'"
        self.run_command(command, "Failed to generate CA CSR")

        with open("esm-ca.cnf", "w") as ca_cnf:
            ca_cnf.write(root_cnf)

        # Self Sign the CA certifcate
        command = "openssl x509 -req  -days 3650  -in esm-ca.csr -signkey " \
                  "esm-ca.key -sha256 -out esm-ca.crt -extfile esm-ca.cnf " \
                  "-extensions root_ca"
        self.run_command(command, "Failed to sign CA certificate")

        # Generate server Key
        command = "openssl genrsa -out server.key 4096"
        self.run_command(command, "Failed to generate server key")

        command = "openssl req -new -key server.key -out server.csr -sha256 " \
                  "-subj '/O=Intel Corp/CN={0}'".format(self.hostname)
        self.run_command(command, "Failed to sign server certificate")

        server_cnf = "{0}\n[alt_names]".format(server_cnf)
        for i in range(len(SERVER_HOSTNAMES)):
            server_cnf = "{0}\nDNS.{1} = {2}".format(server_cnf, i + 1,
                                                     SERVER_HOSTNAMES[i])
        server_cnf = "{0}\nDNS.{1} = {2}".format(server_cnf, i + 2, self.hostname)
        server_cnf = "{0}\nIP.1 = {1}".format(server_cnf, self.ip_address)
        server_cnf = "{0}\nIP.2 = 127.0.0.1".format(server_cnf)
        server_cnf = "{0}\nIP.3 = 192.168.10.1".format(server_cnf)

        with open("server.cnf", "w") as cnf_file:
            cnf_file.write(server_cnf)
        # Sign server certificate using the CA
        command = "openssl x509 -req -days 750 -in server.csr -sha256 " \
                  "-CA esm-ca.crt -CAkey esm-ca.key -CAcreateserial " \
                  "-out server.crt -extfile server.cnf -extensions server"
        self.run_command(command, "Failed to sign server certificate by CA")

        dir_path = f"/opt/{MODULE}"
        if not os.path.exists(dir_path):
            self.run_command(f"sudo mkdir {dir_path}",
                             "Failed to create directory: {}".format(dir_path))
        ca_path = os.path.join(dir_path, "esm-ca.crt")
        command = "sudo cp esm-ca.crt {0}".format(ca_path)
        self.run_command(command, "Unable to copy CA certificate")
        pri_key = "server.key"
        pub_key = "server.crt"
        dir_path = f"/etc/{MODULE}"
        if not os.path.exists(dir_path):
            self.run_command(f"sudo mkdir {dir_path}",
                             f"Failed to create directory: {dir_path}")

        os.environ["KEY_FILE"] = os.path.join(dir_path, pri_key)
        os.environ["CERT_FILE"] = os.path.join(dir_path, pub_key)
        command = f"sudo cp {pri_key} {dir_path}"
        self.run_command(command, "Unable to copy key")
        command = f"sudo cp {pub_key} {dir_path}"
        self.run_command(command, "Unable to copy certificate")
        command = f"sudo chmod 400 {dir_path}/{pri_key}"
        self.run_command(command, "Unable to update certificate permission")
        os.chdir(self.module_path)

    def install_ca_crt(self):
        """
        Install the CA certificate on the build node
        :return:
        """
        self.log.info("Installing Edge Software Configurator CA on the node")
        ca_path = "/usr/local/share/ca-certificates"
        if not os.path.isdir(ca_path):
            self.run_command("sudo mkdir {0}".format(ca_path),
                             "Unable to create directory {0}".format(ca_path))
        command = "sudo cp setup/certs/esm-ca.crt {0}/esm-ca.crt" \
            .format(ca_path)
        self.run_command(command, "Unable to copy CA certificate")
        self.run_command("sudo update-ca-certificates --fresh",
                         "Unable to update  CA certificate")

    def install_docker_ca_crt(self):
        """
        Install the CA certificate into docker directory on the build node
        :return:
        """
        self.log.info("Configuring Edge Software Configurator CA for Docker")
        docker_ca_path = "/etc/docker/certs.d/{0}:{1}".format(self.hostname,
                                                              REGISTRY_PORT)
        if not os.path.isdir(docker_ca_path):
            command = "sudo mkdir -p {0}".format(docker_ca_path)
            self.run_command(command,
                             "Unable to create directory {0}".format(docker_ca_path))
        command = "sudo cp setup/certs/esm-ca.crt {0}/esm-ca.crt" \
            .format(docker_ca_path)
        self.run_command(command, "Unable to copy CA certificate")

    def generate_vmserver_secret(self):
        """
        Generate Secret containing VM server key and store it in VMSERVER_TOKEN
        :return: None
        """
        global VMSERVER_TOKEN
        self.log.info("Generating the VM server secret")
        vmserver_key = self.check_command_output("openssl rand -base64 32")
        vmserver_key = vmserver_key.decode("utf-8")
        for line in vmserver_key.split("\n"):
            VMSERVER_TOKEN = "{0}{1}".format(VMSERVER_TOKEN, line)
        self.vmkey = VMSERVER_TOKEN
        command = f"sudo docker secret create {DOCKER_SECRETS['vmserver-key']} -"
        self.run_command(command, "Failed to generate vmserver key",
                         input=VMSERVER_TOKEN.encode())
        return VMSERVER_TOKEN

    def generate_config_file(self, username, password,
                             edge_id=str(uuid.uuid1()), endpoint_id=1):
        """
        Generate esm config file containing default portainer username,
        password and default EndpointID,EdgeID 
        :param username: Username of default account
        :param password: Password of default account
        :param endpoint_id: Default Endpoint ID
        :param edge_id: Edge ID of default Endpoint
        :return: None
        """
        self.agent_uuid = edge_id
        self.log.info("Generating the Edge Software Configurator server config")
        file_lines = []
        file_lines.append("[USER]")
        file_lines.append("username = {}".format(username))
        file_lines.append("password = {}".format(password))
        file_lines.append("[ENDPOINT]")
        file_lines.append("id = {}".format(endpoint_id))
        file_lines.append("edge_id = {}".format(edge_id))
        data = "\n".join(file_lines)
        command = f"sudo docker secret create {DOCKER_SECRETS['esm-conf']} -"
        self.run_command(command, "Failed to generate esm configuration",
                         input=data.encode())

    def generate_file_secret(self, secret_name, file_name):
        command = f"sudo docker secret create {secret_name} {file_name}"
        self.run_command(command, f"Failed to generate {secret_name} secret")

    def init_amt_setup(self):
        """
        Cloning MPS and RPS Servers from git Repo
        """
        self.log.info("Cloning AMT repo files")
        self.remove_directory("/opt/amt-cert")
        os.chdir(os.path.join(self.module_path, "setup"))

        if os.path.isdir(os.path.join(self.module_path, "setup/open-amt-cloud-toolkit")):
            self.remove_directory(amt_dir)
        command = f"git clone --recursive --depth 1 --branch {AMT_REPO_TAG} {AMT_REPO}"
        self.run_command(command, "Failed to clone AMT repos", mandatory=False)

        try:
            os.chdir(os.path.join(self.module_path, "setup/open-amt-cloud-toolkit/"))
            command = "sudo cp ./.env.template ./.env"
            self.run_command(command, "Failed to copy .env file", mandatory=False)
            command = f"sudo sed -i 's/localhost/{self.ip_address}/' ./.env"
            self.run_command(command, "Failed to update .env file", mandatory=False)
            yaml = YAML(typ='rt')
            yaml.preserve_quotes = True
            data = yaml.load(Path("docker-compose.yml"))
            data['services']['rps']['environment']['no_proxy'] = "localhost,127.0.0.1,vault,$no_proxy"
            data['services']['rps']['environment']['NO_PROXY'] = "localhost,127.0.0.1,vault,$no_proxy"
            data['services']['mps']['environment']['no_proxy'] = "localhost,127.0.0.1,vault,$no_proxy"
            data['services']['mps']['environment']['NO_PROXY'] = "localhost,127.0.0.1,vault,$no_proxy"
            data['services']['vault']['environment']['no_proxy'] = "$no_proxy"
            data['services']['vault']['environment']['NO_PROXY'] = "$no_proxy"
            data['services']['mps']['restart'] = "always"
            data['services']['rps']['restart'] = "always"
            data['services']['vault']['restart'] = "always"
            data['services']['mps'].update({'volumes':['/opt/amt-cert:/mps-microservice/private']})
            data['services']['rps'].update({'volumes':['/opt/amt-cert:/rcs-microservice/private']})
            data['services']['rps'].update({'labels':['com.intel.app=esm']})
            data['services']['mps'].update({'labels':['com.intel.app=esm']})
            data['services']['vault'].update({'labels':['com.intel.app=esm']})
            data['services']['rpsdb'].update({'labels':['com.intel.app=esm']})
            data['services']['webui'].update({'labels':['com.intel.app=esm']})

            del data['services']['vault']['ports']
            del data['services']['rpsdb']['ports']
            del data['services']['webui']['ports']
            yaml.dump(data, Path("docker-compose.yml"))
        except Exception as e:
            self.log.error(
                "Service is not present in docker-compose.yml with error {}".format(e))
        finally:
            os.chdir(CURRENT_DIR)

    def reload_daemon(self):
        """
        Utility to reload and restart the docker daemon
        """
        self.run_command("sudo systemctl daemon-reload",
                         "Failed to reload system daemon")
        self.run_command("sudo systemctl restart docker",
                         "Failed to restart docker service")

    def update_docker_no_proxy(self):
        """
        Update NO_PROXY in docker proxy files
        """
        self.log.info("Updating no proxy in docker proxy files")
        http_file = DOCKER_PROXY_DIR + "/" + HTTP_PROXY_FILE
        https_file = DOCKER_PROXY_DIR + "/" + HTTPS_PROXY_FILE

        def set_proxy(proxy_file):
            try:
                update_no_proxy = False
                have_section = False
                have_service = False
                with open(proxy_file) as config_file:
                    config = configparser.ConfigParser(strict=False)
                    config.read_file(config_file)
                    if config.has_section("Service"):
                        have_service = True
                        no_proxy = config.get("Service", "Environment", fallback=None)
                        for key in config["Service"]:
                            if key == "Environment" or key == "environment" \
                                    and "NO_PROXY" in config["Service"][key]:
                                have_section = True
                                out = re.search("NO_PROXY.*{0}[,\"]"
                                                .format(self.hostname),
                                                config["Service"][key])
                                if not out:
                                    update_no_proxy = True

                    if update_no_proxy:
                        command = f"sudo sed -i 's/NO_PROXY=/NO_PROXY={self.ip_address},{self.hostname},/' {proxy_file}"
                        self.run_command(command,
                                         "Failed to set NO_PROXY in {0}".format(proxy_file),
                                         mandatory=False)
                        self.reload_daemon()
                    elif not have_section and have_service:
                        no_proxy = f"Environment=\"NO_PROXY=127.0.0.1,localhost,{self.ip_address},{self.hostname}\""
                        command = f"sudo tee -a {proxy_file}"
                        self.run_command(command,
                                         "Failed to set NO_PROXY in {0}".format(
                                             proxy_file),
                                         input=no_proxy.encode(),
                                         mandatory=False)
                        self.reload_daemon()

            except FileNotFoundError:
                self.log.info("{0} File does not exist".format(proxy_file))

        set_proxy(http_file)
        set_proxy(https_file)

    def volume_check(self, volume_name):
        """
        Check if volume exists or not
        """
        command = "sudo docker volume ls --format {{.Name}}"
        status = self.run_command(command, get_output=True)
        if status.returncode:
            self.log.error("Failed to check the docker volumes")
            return False
        output = status.stdout.decode('utf-8').split("\n")
        if volume_name in output:
            return True
        else:
            return False

    def esm_login(self):
        """
        Authenticate uer to get jwt and vmkey
        """
        self.log.info("Authenticating User")
        headers = {"Content-Type": "application/json"}
        data = {"username": self.user, "password": self.password}
        res = requests.post("{0}/api/auth".format(self.base_url),
                            headers=headers,
                            json=data,
                            verify=CA_BUNDLE)
        if res.status_code == UNPROCESSABLE_ENTITY:
            return "Unauthorized", ""
        elif res.status_code == SUCCESS:
            res_data = json.loads(res.text)
            return res_data["jwt"], res_data["vmkey"]
        else:
            self.log.error("Authentication failed in Edge Software Configurator server")
            self.log.error(f"{res.status_code}:{str(res.text)}")
            return "", ""

    def create_proxy_config(self):
        """
        Creates proxy config in /opt/esm
        """
        dir_path = f"/opt/{MODULE}"
        if not os.path.exists(dir_path):
            self.run_command(f"sudo mkdir {dir_path}",
                             "Failed to create directory: {}".format(dir_path))

        proxy_file = f"{dir_path}/proxy.conf"
        env_proxy = urllib.request.getproxies()
        self.log.info("Extracting proxy settings from system environment")
        proxy_data = []
        http_proxy = env_proxy.get("http", "")
        https_proxy = env_proxy.get("https", "")

        if http_proxy:
            self.log.info("Proxy settings found.")
            proxy_data.append(f"export http_proxy={http_proxy}")
            proxy_data.append(f"export HTTP_PROXY={http_proxy}")
            if not https_proxy:
                https_proxy = http_proxy
        if https_proxy:
            proxy_data.append(f"export https_proxy={https_proxy}")
            proxy_data.append(f"export HTTPS_PROXY={https_proxy}")

        if proxy_data:
            proxy_data = "\n".join(proxy_data)
            command = f"sudo tee -a {proxy_file}"
            self.run_command(command, "Failed to set PROXY in {0}".format(proxy_file),
                             input=proxy_data.encode(), mandatory=False)

    def stop_agents(self):
        """
        Stop the agents.
        """
        command = "sudo docker-compose -p {0} stop".format(STACK)
        self.run_command(command, "Failed to stop the Edge Software Configurator agents", mandatory=False)
        command = f"sudo docker rm --force {STACK}-collectd"
        self.run_command(command, "Failed to stop the Edge Software Configurator agents", mandatory=False)

    def update_portainer_setting(self, username, password):
        """
        Update the portainer edge agent poll time to 30 sec
        """
        self.log.info("Updating Portainer Setting")
        headers = {"Content-Type": "application/json"}
        data = {"username": username, "password": password}
        res = requests.post("{0}/api/auth".format(self.portainer_url),
                            headers=headers,
                            json=data,
                            verify=CA_BUNDLE)
        if res.status_code != SUCCESS:
            self.log.warn(f"Failed to authenticate: {res.text}")
        else:
            res_data = json.loads(res.text)
            headers = {"Content-Type": "application/json",
                       "Authorization": "Bearer {}".format(res_data["jwt"])}
            data = {"LogoURL": "",
                   "BlackListedLabels": [{"name": "com.intel.app", "value": "esm"}],
                   "AuthenticationMethod": 1,
                   "LDAPSettings": {"AnonymousMode": True,"ReaderDN": "",
                      "URL": "","TLSConfig": {"TLS":False, "TLSSkipVerify": False},
                      "StartTLS": False, "SearchSettings": [{"BaseDN": "", "Filter": "",
                                                          "UserNameAttribute": ""}],
                      "GroupSearchSettings": [{
                            "GroupBaseDN": "",
                            "GroupFilter": "",
                            "GroupAttribute": ""
                         }],
                      "AutoCreateUsers": True
                   },
                   "OAuthSettings": {
                      "ClientID": "",
                      "AccessTokenURI": "",
                      "AuthorizationURI": "",
                      "ResourceURI": "",
                      "RedirectURI": "",
                      "UserIdentifier": "",
                      "Scopes": "",
                      "OAuthAutoCreateUsers": False,
                      "DefaultTeamID": 0
                   },
                   "AllowBindMountsForRegularUsers": True,
                   "AllowPrivilegedModeForRegularUsers": True,
                   "AllowVolumeBrowserForRegularUsers": False,
                   "SnapshotInterval": "5m",
                   "TemplatesURL": "",
                   "EnableHostManagementFeatures": False,
                   "EdgeAgentCheckinInterval": 15,
                   "EnableEdgeComputeFeatures": False,
                   "AllowStackManagementForRegularUsers": True,
                   "AllowHostNamespaceForRegularUsers": True,
                   "AllowDeviceMappingForRegularUsers": True
                    }
            res = requests.put("{0}/api/settings".format(self.portainer_url),
                               headers=headers,
                               json=data,
                               verify=CA_BUNDLE)
            if res.status_code != SUCCESS:
                self.log.warn(f"Failed to update Portainer setting: {res.text}")
            else:
                self.log.info("Successfully updated portainer settings")

    def install_wlm_service(self):
        """
        Start the installation of wlm systemd service
        """
        os.chdir(self.module_path)
        self.log.info("Installing WLM")
        wlm_source_path = os.path.join("setup", "wlm")
        wlm_dest_path = os.path.join("/opt/esm", "wlm")
        if os.path.exists(wlm_dest_path):
            command = f"sudo rm -r {wlm_dest_path}"
            status = self.run_command(command, "Failed to remove old wlm files",
                                      mandatory=False)

        command = f"sudo cp -r {wlm_source_path} {wlm_dest_path}"
        self.run_command(command, "Failed to copy wlm files")
        wlm_script = "setup.sh"

        os.chdir(wlm_dest_path)
        command = f"sudo chmod a+x {wlm_script}"
        self.run_command(command, get_output=True)
        os.environ["WLM_PATH"] = wlm_dest_path
        command = f"sudo -E bash {wlm_script} install"
        self.run_command(command, "Failed to install WLM")
        os.chdir(self.module_path)
        self.log.info("WLM installed")

    def pre_setup(self):
        with progressbar(length=12, fill_char=".", empty_char=" ", width=50,
                         label="Setting up Edge Software Configurator".ljust(
                             self.str_len, " "),
                         show_eta=True, show_percent=True,
                         bar_template="%(label)s [%(bar)s] %(info)s") as pbar:
            time.sleep(1)
            pbar.update(1)
            self.init_swarm()
            pbar.update(1)
            if self.verify_post_install():
                self.reinstallation = True
            else:
                self.remove_stack(uninstall=True, print_logs=False)
            pbar.update(1)
            os.chdir(os.path.join(self.module_path, "setup"))
            self.create_proxy_config()
            if not self.reinstallation:
                self.update_docker_no_proxy()
                pbar.update(1)
                self.generate_certificates()
                pbar.update(1)
                self.install_ca_crt()
                pbar.update(1)
                self.install_docker_ca_crt()
                pbar.update(1)
                self.generate_vmserver_secret()
                pbar.update(1)
                self.generate_config_file(self.username, self.password)
                pbar.update(1)
                os.chdir(os.path.join(self.module_path, "setup"))
                self.generate_file_secret(DOCKER_SECRETS["server.crt"],
                                          "certs/server.crt")
                pbar.update(1)
                self.generate_file_secret(DOCKER_SECRETS["server.key"],
                                          "certs/server.key")
                pbar.update(1)
            self.collectd_config()
            pbar.update(15)

    def display_values(self, name, details):
        print("="*95)
        print("Summary for bare metal provisioning plugin installation")
        print("="*95)
        print("\nNetwork Mode:\t\t\t", name)
        if name == "Dual Ethernet":
            print("Gateway Interface - Upstream:\t", details['UPSTREAM_INTERFACE'])
            print("DHCP Interface - Downstream:\t", details['DOWNSTREAM_INTERFACE'])
            print("\nDHCP Server  (Downstream)  parameters\n")
            print("DHCP Range:\t\t\t", details["DHCP_RANGE"])
            print("DHCP Server Name:\t\t", details['DHCPSERVERNAME'])
            print("DNS_IPs:\t\t\t", details['DNS_IP'])
        if details.get("http_proxy", "") != "":
            print("\nProxy Parameters")
            print("\nhttp_proxy:\t\t\t", details["http_proxy"])
            print("\nhttps_proxy:\t\t\t", details["https_proxy"])
        print("="*95)

    def get_proxy_from_user(self):
        while True:
            custom_proxy = input("\n\nDo you want to customize proxy parameters? "
                                 "Press 'Y' to customize or 'N' to skip\n")
            if repr(custom_proxy) == repr("\x05"):
                raise EdgeSoftwareManagerException("Re-enter the values")
            if custom_proxy == 'Y' or custom_proxy == 'y':
                print("_" * 95)
                print("\nProxy Parameters")
                http_proxy_bm = input("\nEnter http_proxy (Ex: proxy.company.com:121)")
                if repr(http_proxy_bm) == repr("\x05"):
                    raise EdgeSoftwareManagerException("Re-enter the values")
                http_proxy_bm = http_proxy_bm.strip()
                cprint(http_proxy_bm, "cyan")
                https_proxy_bm = input("\nEnter https_proxy (Ex: proxy.company.com:122)")
                if repr(https_proxy_bm) == repr("\x05"):
                    raise EdgeSoftwareManagerException("Re-enter the values")
                https_proxy_bm = https_proxy_bm.strip()
                cprint(https_proxy_bm, "cyan")
                break
            elif custom_proxy == 'N' or custom_proxy == 'n':
                env_proxy = urllib.request.getproxies()
                self.log.info("Extracting proxy settings from system environment")
                http_proxy_bm = env_proxy.get("http", "")
                https_proxy_bm = env_proxy.get("https", "")
                break
            else:
                print("Invalid Option")
        return http_proxy_bm, https_proxy_bm

    def dual_ethernet(self, interfaces):
        my_dict = {}
        for i in interfaces:
            output = self.check_command_output(f"ip addr show {i}")
            try:
                output = self.check_command_output("grep 'inet'", input=output)
                output = self.check_command_output("awk '{print $2}'", input=output)
                output = self.check_command_output("cut -d/ -f1", input=output).decode().split()
                if len(output) != 0:
                    my_dict.update({i: output[0]})
                else:
                    my_dict.update({i: "NO IP"})
            except CalledProcessError:
                my_dict.update({i: "NO IP"})
        print("You selected {0}".format(colored("Dual Ethernet\n\n", "cyan")))
        print("You can select the Gateway and DHCP Interfaces listed below:")
        for i, (k, v) in enumerate(my_dict.items(), start=1):
            print(i, k, " : ", v)
        while True:
            upstreamValue = input(f"\nEnter your choice for Gateway Interface - "
                                  f"Upstream (a number between 1 to {len(interfaces)}): ")
            if repr(upstreamValue) == repr("\x05"):
                raise EdgeSoftwareManagerException("Re-enter the values")
            if not upstreamValue.isnumeric():
                print("Invalid value entered!\n")
            elif int(upstreamValue) > len(interfaces) or int(upstreamValue) < 1:
                print("Invalid value entered!")
            else:
                break
        print(f"You selected {colored(interfaces[int(upstreamValue) - 1], 'cyan')}\n")
        while True:
            downstreamValue = input(f"Enter DHCP Interface - Downstream (a number "
                                    f"between 1 to {len(interfaces)}, except {upstreamValue}): ")
            if repr(downstreamValue) == repr("\x05"):
                raise EdgeSoftwareManagerException("Re-enter the values")
            if not downstreamValue.isnumeric():
                print("Invalid value entered!\n")
            elif downstreamValue == upstreamValue:
                cprint("Error: Invalid value entered! Gateway and DHCP Interface values cannot be the same.\n","red")
            elif int(downstreamValue) > len(interfaces) or int(downstreamValue) < 1:
                print("Invalid value entered!\n")
            else:
                break

        print("You selected {0}\n".format(colored(interfaces[int(downstreamValue) - 1], "cyan")))
        print("="*95)
        print("\nDHCP Server (Downstream) parameters\n")
        try:
            # cidr_value = 2**(32-int(dhcp_range.rpartition('/')[-1]))
            # dhcp_range_max = dhcp_range.replace(dhcp_range.split(".")[-1], str(cidr_value-2))
            # dhcp_range_min = dhcp_range.replace(dhcp_range.split(".")[-1], str(2))
            while True:
                dhcp_range = input("Enter DHCP range (Default value: 192.168.10.0/24): ")
                if dhcp_range == "":
                    dhcp_range = "192.168.10.0/24"
                if repr(dhcp_range) == repr("\x05"):
                    raise EdgeSoftwareManagerException("Re-enter the values")
                cprint(dhcp_range, "cyan")
                try:
                    ip_list = ipaddress.ip_network(dhcp_range, False)
                    ip_list = list(ip_list)
                    dhcp_range_max = str(ip_list[-3])
                    dhcp_range_min = str(ip_list[2])
                    dhcp_broadcast_ip = str(ip_list[-1])
                    dhcp_server_ip = str(ip_list[1])
                    break
                except ValueError as err:
                    print(err)

            dhcp_server_name = input("\nEnter DHCP Server Name (Default "
                                     f"value: {self.hostname}): ")
            dhcp_server_name = dhcp_server_name.strip()
            if repr(dhcp_server_name) == repr("\x05"):
                raise EdgeSoftwareManagerException("Re-enter the values")
            if dhcp_server_name == "":
                dhcp_server_name = self.hostname
            cprint(dhcp_server_name, "cyan")
            print("\nEnter DNS IPs")
            print("Format: <IP_address>,<IP_address>")
            print("Example:122.23.2.1,1.2.3.4,122.23.22.5")
            print("Note: 8.8.8.8 and 8.8.4.4 are already configured")
            try:
                list_dns_ips = [ip for ip in input().split(',')]
                if not all(list_dns_ips):
                    list_dns_ips = [dhcp_server_ip]
                else:
                    for i in list_dns_ips:
                        if bool(ipaddress.ip_address(i)):
                            pass
                        else:
                            print("Invalid IP address")
                    #list_dns_ips.append(dhcp_server_ip)
                    list_dns_ips.insert(0,dhcp_server_ip)
            except ValueError:
                if repr(list_dns_ips) == repr(['\x05']):
                    raise EdgeSoftwareManagerException("Re-enter the values")
                else:
                    print ("Invalid Input")
            dns_ips = ','.join(map(str, list_dns_ips))
            cprint(dns_ips, "cyan")
            http_proxy_bm, https_proxy_bm = self.get_proxy_from_user()
            self.down_interface = interfaces[int(downstreamValue)-1]
            baremetal_details = {"NETWORK_MODE": 1,
                                 "UPSTREAM_INTERFACE": interfaces[int(upstreamValue)-1],
                                 "DOWNSTREAM_INTERFACE": interfaces[int(downstreamValue)-1],
                                 "DHCPRANGEMIN": dhcp_range_min,
                                 "DHCPRANGEMAX": dhcp_range_max,
                                 "DHCPSERVERNAME": dhcp_server_name,
                                 "DHCPSERVERIP": dhcp_server_ip,
                                 "DHCP_BROADCAST_IP": dhcp_broadcast_ip,
                                 "DNS_IP": dns_ips, "DHCP_RANGE": dhcp_range,
                                 "http_proxy": http_proxy_bm,
                                 "https_proxy": https_proxy_bm}
            self.display_values("Dual Ethernet", baremetal_details)
            while True:
                choice = input("Confirm to start the bare metal plugin "
                               "installation, Press 'Y' to confirm or 'R' "
                               "to re-enter values\n")
                if choice == 'Y' or choice == 'y':
                    return baremetal_details
                elif choice == 'R' or choice == 'r':
                    raise EdgeSoftwareManagerException("Re-enter the values")
                else:
                    print("Invalid Option!!")
        except ValueError:
            print("Invalid Input!")
            self.log.error("Error while taking user inputs for Dual ethernet "
                           "BareMetal Installation")

    def single_ethernet(self):
        http_proxy_bm, https_proxy_bm = self.get_proxy_from_user()
        details = {"NETWORK_MODE": 2, "http_proxy": http_proxy_bm,
                   "https_proxy": https_proxy_bm, "DHCPSERVERNAME":self.hostname}
        self.display_values("Single Ethernet", details)
        while True:
            choice = input("Confirm to start the bare metal plugin installation, "
                           "Press 'Y' to confirm or 'R' to re-enter values\n")
            if choice == 'R' or choice == 'r':
                raise EdgeSoftwareManagerException("Re-enter the values")
            elif choice == 'Y' or choice == 'y':
                return details
            else:
                print("Invalid Option")

    def esc_cli(self):
        dual_ethernet_text = """\
        \t\tDHCP server running on the specified downstream interface.The upstream 
        \t\tinterface will serve as a source of internet.In this mode, you can run the ESC
        \t\ton the company/corporate network and also communicate with
        \t\tother devices running on the company/corporate network.
        """
        single_ethernet_text = """\
        \t\tthe router on an isolated network.In this mode, the ESC will not  
        \t\trun any DHCP server
        """
        self.down_interface = ""
        try:
            while True:
                print("=" * 95)
                cprint ("Bare Metal Provisioning  -  Plugin Installation", "blue")
                print("=" * 95)
                print ("Refer documentation to understand more about bare metal "
                    "provisioning.")
                print ("\nNOTE:")
                print ("Press \"Ctrl +E\" to re-enter the values for bare metal "
                    "provisioning plugin installation...\n")
                print ("To start the bare metal provisioning plugin installation, "
                    "select a network mode from the below listed options.\n")
                print(f"{colored('1. Dual Ethernet', 'magenta')} \t With the "
                    f"network mode, you will create an isolated network with the")
                print(dual_ethernet_text)
                print(f"{colored('2. Single Ethernet', 'magenta')} \t With the "
                    f"network mode, you will need to run the ESC setup behind")
                print(single_ethernet_text, "\n")
                print(f"DO NOT select \'{colored('2. Single Ethernet', 'magenta')}\' "
                    f"mode if on company/corporate network.\n")
                command = "find /sys/class/net -mindepth 1 -maxdepth 1 -lname  '*virtual*' -prune -o -printf '%f\n'"
                interfaces = self.check_command_output(command).decode().split()
                if len(interfaces) == 1:
                    print("Your system has only one physical interface.")
                    print(f"Hence the selected mode is {colored('Single Ethernet','cyan')}")
                    details = self.single_ethernet()
                    return details
                else:
                    option = input("Select mode (1 or 2, Default value is Dual Ethernet): ")
                    if repr(option) == repr("\x05"):
                        raise EdgeSoftwareManagerException("Re-enter the values")
                    if option == "1" or option == "":
                        details = self.dual_ethernet(interfaces)
                        return details
                    if option == "2":
                        print("You selected {0}".format(colored("Single Ethernet\n\n", "cyan")))
                        details = self.single_ethernet()
                        return details
                    else:
                        print("Invalid Option")
        except EdgeSoftwareManagerException:
            return self.esc_cli()       

    def _uninstall_plugin(self, plugin_id):
        """
        Uninstall a plugin given its uuid
        """
        headers = {"Authorization": self.jwt}
        res = requests.delete(f"{self.base_url}/api/plugin/{plugin_id}",
                              headers=headers,
                              verify=CA_BUNDLE)
        return res

    def _install_plugin(self, data):
        """
        Install a plugin
        """
        headers = {"Content-Type": "application/json",
                   "Authorization": self.jwt}
        res = requests.post(f"{self.base_url}/api/plugin",
                            headers=headers,
                            json=data,
                            verify=CA_BUNDLE)
        return res

    def _get_plugin_data(self, plugin_id):
        """
        Get the plugin details
        """
        headers = {"Authorization": self.jwt}
        res = requests.get(f"{self.base_url}/api/plugin/{plugin_id}",
                           headers=headers,
                           verify=CA_BUNDLE)
        return res

    def uninstall_bm_plugin(self):

        os.environ["no_proxy"] = "{0},{1},127.0.0.1,localhost,{2}".format(
            self.ip_address, self.hostname, os.environ.get("no_proxy", ""))
        bm_plugin_id = "9d55a8e0-d5f6-443f-99d7-789ffd52ecb8"
        for i in range(3):
            self.user       = getpass.getuser()
            self.password   = getpass.getpass(
                prompt=f"password for {self.user}: ")
            jwt, vmkey = self.esm_login()
            if jwt:
                if jwt == "Unauthorized":
                    print("Sorry, try again.")
                    continue
                self.jwt = jwt
                break
            else:
                print("Failed to authenticate User. Reinstallation Failed.")
        if not self.jwt:
            return Status.FAILED

        res = self._uninstall_plugin(bm_plugin_id)
        if res.status_code == SUCCESS:
            self.log.info("BareMetal Plugin Uninstallation Succeeded.")
            print("BareMetal Plugin Uninstallation Succeeded.")

            return Status.SUCCESS
        else:
            response_data = res.json()

            self.log.error(f"BareMetal Plugin Uninstallation Failed. "
                           f"Details={response_data}")
            print("BareMetal Plugin Uninstallation Failed.")

            return Status.FAILED

    def install_bm_plugin(self, id, data):
        while True:
            res = self._install_plugin(data)
            response_data = res.json()
            if response_data["message"] == "Success":
                print("\nPlugin Installation takes 5-10 minutes.")
                self.log.info("BareMetal Plugin Installation Started.")
                print("BareMetal Plugin Installation Started.")
                print("Installing..", end="")
                while True:
                    self.log.info("Installing BareMetal Plugin")
                    sys.stdout.flush()
                    print(".", end="")
                    sys.stdout.flush()
                    time.sleep(60)
                    res_query = self._get_plugin_data(id)
                    query_data = res_query.json()
                    if query_data["status"] == "Success":
                        # print(query_data["message"])
                        return True
                    elif query_data["status"] == "Failed":
                        print("-Installation Failed.")
                        print("Error:", query_data["message"])
                        del_req = self._uninstall_plugin(id)
                        del_status = del_req.json()
                        if del_status["message"] == "Success":
                            print("-Uninstalled BareMetal Plugin")
                            while True:
                                reinstall = input("Do you want to reinstall with new values?? "
                                                  "Press 'Y' to confirm and 'Q' to quit installation!!")
                                if reinstall == "Y" or reinstall == "y":
                                    details = self.esc_cli()
                                    status = self.baremetal_installation(details)
                                    return status
                                elif reinstall == "Q" or reinstall == "q":
                                    print("Quitting Plugin Installation")
                                    return False
                                else:
                                    print("Invalid Option!!")
            elif response_data["message"] == "Failure":
                print("Error:", response_data["details"])
                self.log.error("Error:", response_data["details"])
                return False

    def baremetal_installation(self, details):
        baremetal_id = "9d55a8e0-d5f6-443f-99d7-789ffd52ecb8"
        bridge_name_dual = "escbr1"
        if details["NETWORK_MODE"] == 1:
            bm_data = {"plugin_id": baremetal_id,
                       "input": [{"name": "NW_DNS_PRIMARY", "value": "8.8.4.4"},
                                 {"name": "NW_DNS_SECONDARY","value": "8.8.8.8"},
                                 {"name": "BRIDGE","value": bridge_name_dual}
                                ]}
        else:
            bm_data = {"plugin_id": baremetal_id, "input": []}
        for k, v in details.items():
            bm_data["input"].append({"name": k, "value": v})
        bm_status = self.install_bm_plugin(baremetal_id, bm_data)
        if bm_status:
            return True
        else:
            return False

    def install_vmagent_plugin(self):
        plugin_id = "158d9b46-f4fa-4055-b81e-cc275a94be4e"
        data = {"plugin_id": plugin_id,
                "input": [{"name": "dhcp_interface", "value": self.down_interface},
                          {"name": "idv", "value": True}]
                }

        while True:
            res = self._install_plugin(data)
            response_data = res.json()
            if response_data["message"] == "Success":
                self.log.info("Virtual Machine Agent Installation Started.")
                print("Virtual Machine Agent Installation Started.")
                print("Installing..", end="")
                while True:
                    self.log.info("Installing Virtual Machine Agent")
                    sys.stdout.flush()
                    print(".", end="")
                    sys.stdout.flush()
                    time.sleep(30)
                    res_query = self._get_plugin_data(plugin_id)
                    query_data = res_query.json()
                    if query_data["status"] == "Success":
                        # print(query_data["message"])
                        return True
                    elif query_data["status"] == "Failed":
                        print("-Installation Failed.")
                        print("Error:", query_data["message"])
                        del_req = self._uninstall_plugin(plugin_id)
                        del_status = del_req.json()
                        if del_status["message"] == "Success":
                            print("-Uninstalled Virtual Machine Agent")
                            while True:
                                reinstall = input(
                                    "Do you want to reinstall Virtual Machine Agent?? "
                                    "Press 'Y' to confirm and 'Q' to quit installation!!")
                                if reinstall == "Y" or reinstall == "y":
                                    status = self.install_vmagent_plugin()
                                    return status
                                elif reinstall == "Q" or reinstall == "q":
                                    print("Quitting Plugin Installation")
                                    return False
                                else:
                                    print("Invalid Option!!")
            elif response_data["message"] == "Failure":
                print("Error:", response_data["details"])
                self.log.error("Error:", response_data["details"])
                return False

    def store_network_details(self, data):
        headers = {"Content-Type": "application/json",
                   "Authorization": self.jwt}
        res = requests.post(f"{self.base_url}/api/settings",
                            headers=headers,
                            json={"network_settings": data},
                            verify=CA_BUNDLE)
        if res.status_code != SUCCESS:
            print("Error:", res)
            raise EdgeSoftwareManagerException("",
                                               "Failed to store Network details",
                                               res.text)
        else:
            print("Successfully stored network setting details.")
            self.log.info("Stored the Network setting details in Database")

    def ping_endpoint(self):
        """
        Ping endpoint agent
        """
        self.log.info("Checking portainer endpoint connection")
        endpoint_id = 1
        url = f"/api/endpoints/{endpoint_id}/docker/_ping"
        headers = {"Authorization": f"Bearer {self.jwt}"}
        res = requests.get(f"{self.portainer_url}{url}",
                           headers=headers,
                           verify=CA_BUNDLE,
                           timeout=60)
        if res.status_code == SUCCESS:
            self.log.info("Portainer endpoint successfully connected")
            return True
        else:
            self.log.warn(f"Failed to connect to the portainer endpoint: {res.text}")
            return False

    def main_install(self):
        """
        Install the components of ESM
        :return: None
        """
        self.log.info("Starting the Edge Software Configurator Installation")
        print("Setup will take approximately 15-20 minutes. "
              "Time may vary depending on the internet connectivity.")
        os.environ["no_proxy"] = "{0},{1},127.0.0.1,localhost,{2}".format(self.ip_address,self.hostname,os.environ.get("no_proxy", ""))
        self.check_docker()
        baremetal_details = self.esc_cli()
        self.username = "admin"
        # Generate random password portainer admin account
        password = self.check_command_output("openssl rand -base64 8").decode("utf-8")
        self.password = password.rstrip()
        self.pre_setup()
        os.chdir(os.path.join(self.module_path, "setup"))
        with progressbar(length=50 + len(AGENT_IMAGES) + RETRIES,
                         fill_char=".", empty_char=" ", width=50,
                         label="Starting services".ljust(self.str_len, " "),
                         show_eta=True, show_percent=True,
                         bar_template="%(label)s [%(bar)s] %(info)s") as pbar:
            time.sleep(1)
            pbar.update(1)
            if not self.reinstallation:
                self.init_amt_setup()
                pbar.update(1)
                self.deploy_amt_setup()
                pbar.update(3)
            if sys_python_ver in PYTHON_VER:
                self.install_wlm_service()
            pbar.update(1)
            os.chdir(os.path.join(self.module_path, "setup"))

            gotify_generated_pasword = ''.join([random.choice(
                string.ascii_letters + string.digits) for
                n in
                range(12)])

            # Deploy gotify
            self.deploy_gotify(gotify_generated_pasword)
            self.log.info("Waiting for Gotify services")
            time.sleep(60)
            gotify_app_token = None
            gotify_client_token = None
            progress = 0

            if self.reinstallation:
                ca_path = CA_BUNDLE
            else:
                ca_path = os.path.join(self.module_path,
                                       "setup/certs/esm-ca.crt")

            # verify if gotify is running
            for i in range(RETRIES):
                # not found, not started
                services1, services2 = self.verify_services(["esm_gotify", "esm_gotify_db"])

                # generate app and client token only in case of fresh installation
                if not self.reinstallation:
                    if not services1 and not services2:
                        # Call POST API to register app
                        gotify_app_token    = self.register_app(ca_path,
                                                    gotify_generated_pasword)
                        gotify_client_token = self.register_client(ca_path,
                                                    gotify_generated_pasword)
                        break

                time.sleep(5)
                progress += 1
                pbar.update(1)

            # generate token secrets only in case of fresh installation
            if not self.reinstallation:
                self.gotify_create_token_secrets(gotify_app_token,
                                                 gotify_client_token)

            if self.reinstallation:
                self.stop_agents()
            pbar.update(1)
            self.deploy_esm_ui()
            pbar.update(3)
            ui_services = SERVICES.copy()
            agent_service = STACK + "_portainer_edge_agent"
            try:
                ui_services.remove(agent_service)
            except ValueError:
                pass
            progress = 0
            for i in range(RETRIES):
                services1, services2 = self.verify_services(ui_services)
                if not services1 and not services2:
                    break
                time.sleep(15)
                progress += 1
                pbar.update(1)
            progress = RETRIES - progress
            pbar.update(progress)
            if services1 or services2:
                services1 = " ".join(services1)
                services2 = " ".join(services2)
                raise EdgeSoftwareManagerException("",
                                                   f"Failed to start the following "
                                                   f"services: {services2} {services1}",
                                                   "")
            self.update_registry()
            pbar.update(len(AGENT_IMAGES))
            self.deploy_esm_agents(pbar, self.username, self.password)
            pbar.update(2)
            # wait until process is finished
            #rni_process.join()
            pbar.update(3)
            os.chdir(self.module_path)
            for i in range(20):
                pbar.update(1)
                time.sleep(1)
            self.update_portainer_setting(self.username, self.password)
            pbar.update(RETRIES+20)
            self.jwt = self.get_token(ca_path, self.username, self.password)
            retries = 5
            while retries != 0:
                if self.ping_endpoint():
                    break
                time.sleep(5)
                retries -= 1
            status = self.baremetal_installation(baremetal_details)
            if status:
                print("\n-{0} Installed BareMetal Plugin".format(colored("Successfully","blue")))
                self.log.info("Successfully installed BareMetal Plugin")
                status = self.install_vmagent_plugin()
                if status:
                    print("\n-{0} Installed Virtual Machine Agent".
                          format(colored("Successfully", "blue")))
                    self.log.info("Successfully installed Virtual Machine "
                                  "Agent")
                else:
                    print("\n-{0} to install Virtual Machine Agent".format(
                        colored("Failed", "red")))
                    self.log.info("Failed to Install Virtual Machine Agent")
                    raise EdgeSoftwareManagerException("",
                                                       "Failed to install Virtual Machine Agent"
                                                       "")
                self.store_network_details(baremetal_details)
            else:
                print("\n-{0} to install BareMetal Plugin".format(colored("Failed","red")))
                self.log.info("Failed to Install BareMetal Plugin")
                raise EdgeSoftwareManagerException("",
                                                   "Failed to install BareMetal Plugin."
                                                   "")

    def verify_services(self, service_list=None):
        """
        Verify if the required services are running
        :return:
        """
        self.log.info("Verifying the required services")
        not_found = []
        not_started = []
        command = "sudo docker service ls --format '{{.Name}} {{.Replicas}}'"
        status = self.run_command(command, get_output=True)
        if status.returncode:
            raise EdgeSoftwareManagerException(status.stdout + status.stderr,
                                               "Failed to verify services",
                                               command)
        output = status.stdout.decode('utf-8').split("\n")
        output = [i.split(" ") for i in output if i]

        if not service_list:
            service_list = SERVICES

        for service in service_list:
            found = False
            for item in output:
                if service == item[0]:
                    found = True
                    try:
                        replica = item[1].split("/")
                        if replica[0] != replica[1]:
                            not_started.append(service)
                    except IndexError:
                        pass
                    break
            if not found:
                not_found.append(service)

        return not_found, not_started

    def verify_containers(self):
        """
        Verify if the required containers are running
        :return:
        """
        self.log.info("Verifying the required containers")
        not_found = []
        for container in SA_CONTAINER:
            command = f"sudo docker ps --filter name={container} " \
                      "--filter status=running"
            status = self.run_command(command, get_output=True)
            if status.returncode:
                raise EdgeSoftwareManagerException(status.stdout+status.stderr,
                                                   "Failed to verify {} container".format(container),
                                                   command)
            output = status.stdout.decode('utf-8').split("\n")
            output = [i for i in output if i]
            exists = False
            if len(output) >= 2:
                for line in output:
                    line = line.split(" ")
                    if container in line:
                        exists = True
                        break
            if not exists:
                not_found.append(container)

        return not_found

    def remove_stack(self, uninstall=False, print_logs=True):
        """
        Remove ESM components
        :param uninstall: Flag to specify if ESM has to be uninstalled
        :return: None
        """
        if not uninstall:
            self.log.error("Setup Failed")
            log_file = os.path.join(self.output_path, "components.log")
            self.log.info(f"For more logs related to failure, please check {log_file}")
            format_text = 40 * "*"
            for container in SA_CONTAINER:
                text = f"\n\n\n{format_text} Logs for container {container} {format_text}\n"
                self.run_command(f"tee -a  {log_file}", input=text.encode())
                with open(log_file, "a+") as _lfile:
                    self.run_command(f"sudo docker logs {container} -t",
                                     stdout=_lfile, stderr=_lfile,
                                     get_output=True)

            for service in SERVICES:
                text = f"\n\n\n{format_text} Logs for service {service} {format_text}\n"
                self.run_command(f"tee -a  {log_file}", input=text.encode())
                with open(log_file, "a+") as _lfile:
                    self.run_command(f"sudo docker service logs {service} -t",
                                     stdout=_lfile, stderr=_lfile,
                                     get_output=True)
            text = f"\n\n\n{format_text} Logs for docker from systemd {format_text}\n"
            self.run_command(f"tee -a  {log_file}", input=text.encode())
            with open(log_file, "a+") as _lfile:
                self.run_command(f"journalctl -u docker --since \"{self.setup_start_time}\"",
                                 stdout=_lfile, stderr=_lfile,
                                 get_output=True)
            # command = f"sudo cp /opt/rni/builder.log {self.output_path}/rni.log"
            # self.run_command(command,
            #                  "Failed to copy RNI logs", mandatory=False)

        if print_logs:
            self.log.info("Removing the setup")
            print("Removing docker container")
            self.log.info("Removing docker container")
        sa_container = " ".join(SA_CONTAINER)
        other_container = " ".join(OTHER_CONTAINERS)
        amt_container = " ".join(AMT_CONTAINER)
        command = "sudo docker stop  {} {} {} ".format(
            sa_container, other_container, amt_container)
        self.run_command(command, get_output=True)
        command = "sudo docker rm  {} {} {} ".format(
            sa_container, other_container, amt_container)
        self.run_command(command, get_output=True)

        if print_logs:
            print("Removing docker stacks")
            self.log.info("Removing docker stacks")
        self.run_command("sudo docker stack rm esm", get_output=True)
        if print_logs:
            print("Removing files")
            self.log.info("Removing files")
        self.uninstall_wlm_service()
        command = "sudo rm -rf  {}".format(" ".join(HOST_FILE_TO_REM))
        self.run_command(command, get_output=True)
        command = "sudo docker network rm android cicaas_default " \
                  "esm_default esm_openamtnetwork"
        self.run_command(command, get_output=True)

        secrets = list(DOCKER_SECRETS.values())
        command = "sudo docker secret rm {}".format(" ".join(secrets))
        self.run_command(command, get_output=True)
        self.remove_containers("recipe_package")
        self.remove_containers("esm-vm-agent")

        if print_logs:
            print("Removing docker volumes")
            self.log.info("Removing docker volumes")
        for i in range(RETRIES):
            retry = False
            for volume in DOCKER_VOLUMES:
                command = "sudo docker volume rm -f {0}".format(volume)
                self.run_command(command, get_output=True)
            command = "sudo docker volume ls  --format {{.Name}}"
            status = self.run_command(command, get_output=True)
            if not status.returncode:
                existing_volumes = status.stdout.decode('utf-8').split("\n")
                existing_volumes = [i for i in existing_volumes if i]
                for volume in DOCKER_VOLUMES:
                    if volume in existing_volumes:
                        retry = True
            if not retry:
                break

            time.sleep(5)

    def kvm_check(self):
        """
        Check KVM acceralation is supported or not.
        """
        self.log.info("Checking KVM acceleration support")
        self.run_command("sudo apt install -y cpu-checker",
                         "Failed to install cpu-checker")
        status = self.run_command("kvm-ok", get_output=True)
        if status.returncode == 1:
            print("{0} KVM acceleration can NOT be used".format(
                colored("WARNING:", "yellow")))
            print("{0} To manage Virtual Machine using Edge Software Configurator, " \
                  "please enable VT-x in the BIOS Setup.".format(
                colored("WARNING:", "yellow")))

    def remove_containers(self, name):
        command = f"sudo docker ps -a --filter name={name} --format 'table {{{{.ID}}}}'"
        status = self.run_command(command, get_output=True)
        if not status.returncode:
            output = status.stdout.decode('utf-8').split("\n")
            for item in output:
                command = f"sudo docker rm --force {item}"
                self.run_command(command, stdout=DEVNULL, stderr=DEVNULL,
                                 get_output=True)

    def remove_images(self):
        """
        Remove the images built by the application
        """
        self.log.info("Removing images")
        print("Removing images")
        images_list = []
        # for image in CIC_IMAGES:
        #     images_list.append(f"{self.ip_address}:{REGISTRY_PORT}/{image}")
        # for image in AGENT_IMAGES:
        #     images_list.append(f"{self.ip_address}:{REGISTRY_PORT}/{image}")
        for image in AGENT_IMAGES:
            images_list.append(f"{self.hostname}:{REGISTRY_PORT}/{image}")

        command = "sudo docker rmi -f {}".format(" ".join(images_list))
        self.run_command(command, stdout=DEVNULL, stderr=STDOUT, get_output=True)
        #images = list(DOCKER_IMAGES.values())
        #command = "sudo docker rmi -f {}".format(" ".join(images))
        self.run_command(command, stdout=DEVNULL, stderr=STDOUT, get_output=True)
        for i in ADDITIONAL_IMAGES:
            command = "sudo docker rmi -f {}".format(i)
            self.run_command(command, stdout=DEVNULL, stderr=STDOUT,
                             get_output=True)

        command = "sudo docker images --filter reference=*recipe_package* " \
                  "--format 'table {{.ID }}'"
        status = self.run_command(command, get_output=True)
        if status.returncode:
            self.log.warn((status.stdout + status.stderr).decode('utf-8'))
            self.log.warn("Failed to get recipe images")
        else:
            output = status.stdout.decode('utf-8').split("\n")
            for item in output:
                command = f"sudo docker rmi --force {item}"
                self.run_command(command, stdout=DEVNULL, stderr=DEVNULL,
                                 mandatory=False)

        command = "sudo docker image prune"
        self.run_command(command, input="y".encode(), stdout=DEVNULL,
                         stderr=STDOUT, get_output=True)

    def main_post_install(self):
        """
        Check if ESM has been properly installed or not
        :return: None
        """
        self.log.info("Verifying the installation")
        with progressbar(length=RETRIES + 2, fill_char=".", empty_char=" ", width=50,
                         label="Verifying Edge Software Configurator".ljust(
                             self.str_len, " "),
                         show_eta=True, show_percent=True,
                         bar_template="%(label)s [%(bar)s] %(info)s") as pbar:
            time.sleep(0.5)
            pbar.update(1)
            result = False
            containers_not_found = []
            service_not_found = []
            service_not_started = []
            for i in range(RETRIES):
                containers_not_found = self.verify_containers()
                service_not_found, service_not_started = self.verify_services()
                if not containers_not_found \
                  and not service_not_found \
                  and not service_not_started:
                    result = True
                    pbar.update(1)
                    break
                pbar.update(1)
                time.sleep(5)
            if not result:
                containers_not_found = " ".join(containers_not_found)
                service_not_found = " ".join(service_not_found)
                service_not_started = " ".join(service_not_started)
                self.log.error("Following components not found: \
                                {0} {1} {2}".format(containers_not_found,
                                                    service_not_found,
                                                    service_not_started))
                raise EdgeSoftwareManagerException("",
                                            "Some components of Edge Software Configurator failed to start successfully: "\
                                            "{0} {1} {2}".format(containers_not_found,
                                                                 service_not_found,
                                                                 service_not_started),
                                            "")
            pbar.update(RETRIES)
        self.kvm_check()
        self.verify_pythonver()
        print(f"Please visit {colored('https://127.0.0.1:3002', 'blue')} "
              f"and login with your Linux user credential to access Edge Software Configurator.")
        print("To learn how to use Edge Software Configurator, please open the Get Started Guide "
              "https://software.intel.com/content/www/us/en/develop/documentation/edge-software-configurator-doc/top.html")

    def verify_post_install(self):
        """
        Verify if the ESM is installed or not
        :return: True if installed else False
        """
        containers = self.verify_containers()
        services1, services2 = self.verify_services()

        if containers or services1 or services2:
            self.log.info("Following components not found: \
                           {0} {1} {2}".format(containers,
                                               services1,
                                               services2))
            return False
        else:
            return True

    def remove_directory(self, directory):
        """
        Remove directory
        """
        command = f"sudo rm -r {directory}"
        status = self.run_command(command, get_output=True)
        if status.returncode:
            return False
        else:
            return True

    def uninstall_wlm_service(self):
        """
        Start the installation of wlm systemd service
        """
        self.log.info("Unstalling WLM")
        wlm_dest_path = os.path.join("/opt/esm", "wlm")

        if os.path.exists(wlm_dest_path):
            os.chdir(wlm_dest_path)
            wlm_script = "setup.sh"
            command = f"sudo -E bash {wlm_script} uninstall"
            status = self.run_command(command, "Failed to remove wlm", mandatory=False)
        os.chdir(self.module_path)
        self.log.info("WLM uninstalled")

    def uninstall(self):
        """
        Uninstall ESM
        :return: True for successful installation
        """
        # rni_dir = "/opt/rni"
        self.uninstall_bm_plugin()
        print(colored("Uninstalling the component Edge Software Configurator",
                      "yellow"))
        self.log.info("Uninstalling the component Edge Software Configurator")
        self.remove_stack(uninstall=True)
        # self.remove_directory(rni_dir)
        os.chdir(os.path.join(self.module_path, "setup"))
        self.remove_directory(amt_dir)
        self.remove_directory("/opt/amt-cert")

        self.remove_images()
        return True

    def cleanup(self):
        """
        Removes temporary files
        :return: None
        """
        try:
            self.log.info("Removing temporary files and directories")
            os.chdir(self.module_path)
            # if os.path.exists('setup'):
            #     shutil.rmtree('setup')
            certs = os.path.join(self.module_path, "setup/certs")
            if os.path.exists(certs):
                os.chdir(certs)
                certs_file = ["esm-ca.cnf", "esm-ca.csr", "esm-ca.key",
                              "esm-ca.srl", "server.cnf", "server.crt",
                              "server.csr", "server.key"]

                command = "sudo rm -rf  {}".format(" ".join(certs_file))
                self.run_command(command, get_output=True)
            amt_dir_path = os.path.join(self.module_path,"setup/open-amt-cloud-toolkit")
            if os.path.exists(amt_dir_path):
                self.log.info("Removing amt dir")
                self.remove_directory(amt_dir_path)
            self.log.info("Completed cleanup")
        except Exception as e:
            print("{0} {1}".format(colored("WARNING:", "yellow"), e))
            self.log.warn("Exception : {0}".format(e))
            self.log.warn("Failed Cleanup")

    def verify_pythonver(self):
        if sys_python_ver not in PYTHON_VER:
           print(f"{colored('WARNING:', 'yellow')}  Currently WLM does not support python"
                 + sys_python_ver +
                 f" version")

def main_install(module_path, output_path):
    """
    Install ESM
    """
    esm_install = EdgeSoftwareManager(module_path, output_path)
    try:
        status = esm_install.install()
    except Exception as e:
        esm_install.log.error(e)
        esm_install.log.clean()
        os.chdir(CURRENT_DIR)
        raise e
    esm_install.log.clean()
    return status


def verify_install(module_path, output_path):
    """
    Verify ESM installation
    """
    esm_install = EdgeSoftwareManager(module_path, output_path)
    try:
        status = esm_install.verify_post_install()
    except Exception as e:
        esm_install.log.error(e)
        status = False
    esm_install.log.clean()
    os.chdir(CURRENT_DIR)
    return status


def main_uninstall(module_path, output_path, _):
    """
    Uninstall ESM
    """
    esm_install = EdgeSoftwareManager(module_path, output_path)
    try:
        status = esm_install.uninstall()
    except Exception as e:
        esm_install.log.error(e)
        status = False
    esm_install.log.clean()
    os.chdir(CURRENT_DIR)
    return status

